import { fileURLToPath } from 'url';
import { useLogger, addTemplate, addServerTemplate, addTypeTemplate, resolveAlias, createResolver, resolveFiles, addPlugin, addServerHandler, addImports, addServerImports, useNitro, defineNuxtModule } from '@nuxt/kit';
import { existsSync, promises } from 'node:fs';
import { relative } from 'pathe';
import { basename } from 'node:path';
import { printSourceLocation, parse, Source, OperationTypeNode } from 'graphql';
import { Generator, FieldNotFoundError, TypeNotFoundError, FragmentNotFoundError } from 'graphql-typescript-deluxe';
import color from 'picocolors';
import { validateGraphQlDocuments } from '@graphql-tools/utils';
import fs from 'node:fs/promises';
import { generate } from '@graphql-codegen/cli';
import * as PluginSchemaAst from '@graphql-codegen/schema-ast';
import { loadSchema } from '@graphql-tools/load';
import { defu } from 'defu';
import * as micromatch from 'micromatch';
import { ConfirmPrompt } from '@clack/core';
import isUnicodeSupported from 'is-unicode-supported';
import { existsSync as existsSync$1 } from 'fs';
import { onDevToolsInitialized, extendServerRpc } from '@nuxt/devtools-kit';

const name = "nuxt-graphql-middleware";
const version = "5.0.0-alpha.9";

const logger = useLogger(name);
const defaultOptions = {
  downloadSchema: true,
  schemaPath: "~~/schema.graphql",
  serverApiPrefix: "/api/graphql_middleware",
  graphqlEndpoint: "",
  debug: false,
  includeComposables: true,
  documents: [],
  devtools: true,
  errorOverlay: true,
  graphqlConfigFilePath: "./graphql.config.ts"
};
function validateOptions(options) {
  if (!options.graphqlEndpoint) {
    throw new Error("Missing graphqlEndpoint.");
  }
}
const fileExists = (path, extensions = ["js", "ts", "mjs"]) => {
  if (!path) {
    return null;
  } else if (existsSync(path)) {
    return path;
  }
  const extension = extensions.find(
    (extension2) => existsSync(`${path}.${extension2}`)
  );
  return extension ? `${path}.${extension}` : null;
};

const SYMBOL_CROSS = "x";
const SYMBOL_CHECK = "\u2714";
function getMaxLengths(entries) {
  let name = 0;
  let path = 0;
  let type = 0;
  for (const entry of entries) {
    if (entry.type.length > type) {
      type = entry.type.length;
    }
    if (entry.name.length > name) {
      name = entry.name.length;
    }
    if (entry.path.length > path) {
      path = entry.path.length;
    }
  }
  return { name, path, type };
}
function logAllEntries(entries) {
  const lengths = getMaxLengths(entries);
  let prevHadError = false;
  for (const entry of entries) {
    const hasErrors = entry.errors.length > 0;
    const icon = hasErrors ? color.red(SYMBOL_CROSS) : color.green(SYMBOL_CHECK);
    const type = entry.type.padEnd(lengths.type);
    const namePadded = color.bold(entry.name.padEnd(lengths.name));
    const name = hasErrors ? color.red(namePadded) : color.green(namePadded);
    const path = color.dim(entry.path);
    const parts = [icon, type, name, path];
    if (hasErrors && !prevHadError) {
      process.stdout.write("-".repeat(process.stdout.columns) + "\n");
    }
    logger.log(parts.join(" | "));
    if (hasErrors) {
      const errorLines = [];
      entry.errors.forEach((error) => {
        let output = color.red(error.message);
        if (error.source && error.locations) {
          for (const location of error.locations) {
            output += "\n\n" + color.red(printSourceLocation(error.source, location));
          }
        }
        errorLines.push(output);
      });
      logger.log(
        errorLines.join("\n").split("\n").map((v) => "    " + v).join("\n")
      );
      process.stdout.write("-".repeat(process.stdout.columns) + "\n");
    }
    prevHadError = hasErrors;
  }
  logger.restoreStd();
}

class CollectedFile {
  filePath;
  fileContents;
  isOnDisk;
  parsed;
  constructor(filePath, fileContents, isOnDisk = false) {
    this.filePath = filePath;
    this.fileContents = fileContents;
    this.isOnDisk = isOnDisk;
    this.parsed = parse(fileContents);
  }
  static async fromFilePath(filePath) {
    const content = (await promises.readFile(filePath)).toString();
    return new CollectedFile(filePath, content, true);
  }
  /**
   * If isOnDisk, re-read file contents from disk, then parse it (syntax only).
   */
  async update() {
    if (this.isOnDisk) {
      const newContents = (await promises.readFile(this.filePath)).toString();
      if (newContents === this.fileContents) {
        return false;
      }
      this.fileContents = newContents;
      this.parsed = parse(newContents);
      return true;
    }
    return false;
  }
}

class Collector {
  constructor(schema, helper) {
    this.schema = schema;
    this.helper = helper;
    const mappedOptions = { ...helper.options.codegenConfig };
    if (!mappedOptions.output) {
      mappedOptions.output = {};
    }
    if (!mappedOptions.output.buildTypeDocFilePath) {
      mappedOptions.output.buildTypeDocFilePath = (filePath) => {
        return this.filePathToBuildRelative(filePath);
      };
    }
    this.generator = new Generator(schema, mappedOptions);
  }
  /**
   * All collected files.
   */
  files = /* @__PURE__ */ new Map();
  /**
   * The code generator.
   */
  generator;
  /**
   * A map of operation name and timestamp when the operation was last validated.
   */
  operationTimestamps = /* @__PURE__ */ new Map();
  /**
   * The generated operations and fragments.
   */
  rpcItems = /* @__PURE__ */ new Map();
  /**
   * The registered templates.
   */
  templates = [];
  /**
   * The generated template contents.
   */
  templateResult = /* @__PURE__ */ new Map();
  async reset() {
    this.files.clear();
    this.generator.reset();
    this.operationTimestamps.clear();
    this.rpcItems.clear();
  }
  async updateSchema(schema) {
    this.schema = schema;
    this.generator.updateSchema(schema);
    await this.reset();
    await this.initDocuments();
  }
  filePathToBuildRelative(filePath) {
    return "./" + this.helper.toBuildRelative(filePath);
  }
  filePathToSourceRelative(filePath) {
    return "./" + relative(process.cwd(), filePath);
  }
  operationToLogEntry(operation, errors) {
    return {
      name: operation.graphqlName,
      type: operation.operationType,
      path: this.filePathToSourceRelative(operation.filePath),
      errors
    };
  }
  getTemplate(template) {
    const content = this.templateResult.get(template);
    if (content === void 0) {
      throw new Error(`Missing template content: ${template}`);
    }
    return content;
  }
  /**
   * Executes code gen and performs validation for operations.
   */
  buildState() {
    const output = this.generator.build();
    const operations = output.getCollectedOperations();
    const generatedCode = output.getGeneratedCode();
    this.templates.forEach((template) => {
      if (template.build) {
        const filename = template.options.path + ".js";
        this.templateResult.set(
          filename,
          template.build(output, this.helper).trim()
        );
      }
      if (template.buildTypes) {
        const filename = template.options.path + ".d.ts";
        this.templateResult.set(
          filename,
          template.buildTypes(output, this.helper).trim()
        );
      }
    });
    const fragmentMap = /* @__PURE__ */ new Map();
    const operationSourceMap = /* @__PURE__ */ new Map();
    for (const code of generatedCode) {
      if (code.type === "fragment" && code.graphqlName && code.source) {
        fragmentMap.set(code.graphqlName, code.source);
      } else if (code.type === "operation" && code.graphqlName && code.source) {
        operationSourceMap.set(code.graphqlName, code.source);
      }
    }
    let hasErrors = false;
    const logEntries = [];
    for (const operation of operations) {
      const previousTimestamp = this.operationTimestamps.get(
        operation.graphqlName
      );
      if (previousTimestamp === operation.timestamp) {
        continue;
      }
      const fragments = operation.getGraphQLFragmentDependencies().map((v) => fragmentMap.get(v) || "").join("\n");
      const fullOperation = operationSourceMap.get(operation.graphqlName) + fragments;
      const source = new Source(fullOperation, basename(operation.filePath));
      const document = parse(source);
      const errors = validateGraphQlDocuments(this.schema, [document]);
      if (errors.length) {
        hasErrors = true;
      } else {
        this.operationTimestamps.set(operation.graphqlName, operation.timestamp);
      }
      const shouldLog = errors.length || !this.helper.options.logOnlyErrors;
      if (shouldLog) {
        logEntries.push(this.operationToLogEntry(operation, errors));
      }
    }
    logAllEntries(logEntries);
    if (hasErrors) {
      throw new Error("GraphQL errors");
    }
    if (this.helper.isDev) {
      for (const code of generatedCode) {
        const id = `${code.identifier}_${code.graphqlName}`;
        if (code.identifier === "fragment" || code.identifier === "mutation" || code.identifier === "query") {
          if (this.rpcItems.get(id)?.timestamp === code.timestamp) {
            continue;
          }
          const fragmentDepdendencies = code.getGraphQLFragmentDependencies().map((name) => fragmentMap.get(name) || "").join("\n\n");
          this.rpcItems.set(id, {
            id,
            timestamp: code.timestamp,
            source: code.source + "\n\n" + fragmentDepdendencies,
            name: code.graphqlName,
            filePath: code.filePath,
            identifier: code.identifier
          });
        }
      }
    }
  }
  buildErrorMessage(error) {
    let output = "";
    if (error instanceof FieldNotFoundError || error instanceof TypeNotFoundError || error instanceof FragmentNotFoundError) {
      const filePath = error.context?.filePath;
      const file = filePath ? this.files.get(filePath) : null;
      if (filePath) {
        output += ` | ${this.filePathToSourceRelative(filePath)}
`;
      }
      output += "\n" + error.message + "\n\n";
      if (file) {
        output += file.fileContents;
      }
    } else if (error instanceof Error) {
      output += "\n" + error.message;
    }
    return output;
  }
  logError(error) {
    let output = `${SYMBOL_CROSS}`;
    output += this.buildErrorMessage(error);
    logger.error(color.red(output));
  }
  /**
   * Initialise the collector.
   *
   * In dev mode, the method will call itself recursively until all documents
   * are valid.
   *
   * If not in dev mode the method will throw an error when documents are not
   * valid.
   */
  async init() {
    try {
      await this.initDocuments();
    } catch {
      if (this.helper.isDev) {
        const shouldRevalidate = await this.helper.prompt.confirm(
          "Do you want to revalidate the GraphQL documents?"
        );
        if (shouldRevalidate === "yes") {
          await this.reset();
          return this.init();
        }
      }
      throw new Error("Graphql document validation failed.");
    }
  }
  /**
   * Initialise the collector.
   */
  async initDocuments() {
    try {
      const files = await this.helper.getImportPatternFiles();
      for (const filePath of files) {
        await this.addFile(filePath);
      }
      const nuxtConfigDocuments = this.helper.options.documents.join("\n\n");
      if (nuxtConfigDocuments.length) {
        const filePath = this.helper.paths.nuxtConfig;
        const file = new CollectedFile(filePath, nuxtConfigDocuments, false);
        this.files.set(filePath, file);
        this.generator.add({
          filePath,
          documentNode: file.parsed
        });
      }
      this.buildState();
      logger.success("All GraphQL documents are valid.");
    } catch (e) {
      this.logError(e);
      throw new Error("GraphQL document validation failed.");
    }
  }
  /**
   * Add a file.
   */
  async addFile(filePath) {
    const file = await CollectedFile.fromFilePath(filePath);
    if (!file.fileContents) {
      return null;
    }
    this.files.set(filePath, file);
    this.generator.add({
      filePath,
      documentNode: file.parsed
    });
    return file;
  }
  async handleAdd(filePath) {
    if (!this.helper.matchesImportPattern(filePath)) {
      return false;
    }
    const result = await this.addFile(filePath);
    return !!result;
  }
  async handleChange(filePath) {
    if (!this.helper.matchesImportPattern(filePath)) {
      return false;
    }
    const file = this.files.get(filePath);
    if (!file) {
      return this.handleAdd(filePath);
    }
    try {
      const needsUpdate = await file.update();
      if (!needsUpdate) {
        return false;
      }
      this.generator.update({
        filePath,
        documentNode: file.parsed
      });
    } catch {
      return this.handleUnlink(filePath);
    }
    return true;
  }
  handleUnlink(filePath) {
    const file = this.files.get(filePath);
    if (!file) {
      return false;
    }
    this.files.delete(filePath);
    this.generator.remove(filePath);
    return true;
  }
  handleUnlinkDir(folderPath) {
    let anyHasChanged = false;
    for (const filePath of [...this.files.keys()]) {
      if (filePath.startsWith(folderPath)) {
        const hasChanged = this.handleUnlink(filePath);
        if (hasChanged) {
          anyHasChanged = true;
        }
      }
    }
    return anyHasChanged;
  }
  /**
   * Handle the watcher event for the given file path.
   */
  async handleWatchEvent(event, filePath) {
    let hasChanged = false;
    const oldOperationTimestamps = new Map(this.operationTimestamps);
    try {
      if (event === "add") {
        hasChanged = await this.handleAdd(filePath);
      } else if (event === "change") {
        hasChanged = await this.handleChange(filePath);
      } else if (event === "unlink") {
        hasChanged = this.handleUnlink(filePath);
      } else if (event === "unlinkDir") {
        hasChanged = this.handleUnlinkDir(filePath);
      }
      if (hasChanged) {
        this.buildState();
      }
    } catch (e) {
      this.generator.resetCaches();
      logger.error("Failed to update GraphQL code.");
      this.logError(e);
      return {
        hasChanged: false,
        affectedOperations: [],
        error: { message: this.buildErrorMessage(e) }
      };
    }
    const affectedOperations = [];
    if (hasChanged) {
      logger.success("Finished GraphQL code update successfully.");
      for (const [name, newTimestamp] of this.operationTimestamps) {
        const oldTimestamp = oldOperationTimestamps.get(name);
        if (!oldTimestamp || oldTimestamp !== newTimestamp) {
          affectedOperations.push(name);
        }
      }
    }
    return { hasChanged, affectedOperations };
  }
  /**
   * Adds a virtual template (not written to disk) for both Nuxt and Nitro.
   *
   * For some reason a template written to disk works for both Nuxt and Nitro,
   * but a virtual template requires adding two templates.
   */
  addVirtualTemplate(template) {
    const filename = template.options.path + ".js";
    const getContents = () => this.getTemplate(filename);
    addTemplate({
      filename,
      getContents
    });
    addServerTemplate({
      // Since this is a virtual template, the name must match the final
      // alias, example:
      // - nuxt-graphql-middleware/foobar.mjs => #nuxt-graphql-middleware/foobar
      //
      // That way we can reference the same template using the alias in both
      // Nuxt and Nitro environments.
      filename: "#" + template.options.path,
      getContents
    });
  }
  /**
   * Adds a template that dependes on Collector state.
   */
  addTemplate(template) {
    this.templates.push(template);
    if (template.build) {
      if (template.options.virtual) {
        this.addVirtualTemplate(template);
      } else {
        const filename = template.options.path + ".js";
        addTemplate({
          filename,
          write: true,
          getContents: () => this.getTemplate(filename)
        });
      }
    }
    if (template.buildTypes) {
      const filename = template.options.path + ".d.ts";
      addTypeTemplate(
        {
          filename,
          write: true,
          getContents: () => this.getTemplate(filename)
        },
        {
          nuxt: true,
          nitro: true
        }
      );
    }
  }
}

class SchemaProvider {
  constructor(helper) {
    this.helper = helper;
  }
  /**
   * The raw schema content.
   */
  schemaContent = "";
  /**
   * The parsed schema object.
   */
  schema = null;
  async init() {
    try {
      await this.loadSchema();
    } catch (error) {
      logger.error(error);
      const hasLoaded = await this.loadFromDiskFallback();
      if (!hasLoaded) {
        throw new Error("Failed to load GraphQL schema.");
      }
    }
  }
  async loadFromDiskFallback() {
    const hasSchemaOnDisk = await this.hasSchemaOnDisk();
    if (this.helper.isDev && hasSchemaOnDisk && this.helper.options.downloadSchema) {
      const shouldUseFromDisk = await this.helper.prompt.confirm(
        "Do you want to continue with the previously downloaded schema from disk?"
      );
      if (shouldUseFromDisk === "yes") {
        await this.loadSchema({ forceDisk: true });
        return true;
      }
    }
    return false;
  }
  /**
   * Loads the schema from disk.
   *
   * @returns The schema contents from disk.
   */
  async loadSchemaFromDisk() {
    const fileExists = await this.hasSchemaOnDisk();
    if (!fileExists) {
      logger.error(
        '"downloadSchema" is set to false but no schema exists at ' + this.helper.paths.schema
      );
      throw new Error("Missing GraphQL schema.");
    }
    logger.info(`Loading GraphQL schema from disk: ${this.helper.paths.schema}`);
    return await fs.readFile(this.helper.paths.schema).then((v) => v.toString());
  }
  /**
   * Downloads the schema and saves it to disk.
   *
   * @returns The schema contents.
   */
  downloadSchema() {
    const endpoint = this.helper.options.graphqlEndpoint;
    if (!endpoint) {
      throw new Error("Missing graphqlEndpoint config.");
    }
    const pluginConfig = this.helper.options.codegenSchemaConfig?.urlSchemaOptions;
    const schemaAstConfig = this.helper.options.codegenSchemaConfig?.schemaAstConfig || {
      sort: true
    };
    const config = {
      schema: endpoint,
      pluginLoader: (name) => {
        switch (name) {
          case "@graphql-codegen/schema-ast":
            return Promise.resolve(PluginSchemaAst);
        }
        throw new Error(`graphql-codegen plugin not found: ${name}`);
      },
      silent: true,
      errorsOnly: true,
      config: pluginConfig,
      generates: {
        [this.helper.paths.schema]: {
          plugins: ["schema-ast"],
          config: schemaAstConfig
        }
      }
    };
    logger.info(`Downloading GraphQL schema from "${endpoint}".`);
    return generate(config, true).then((v) => v[0]?.content);
  }
  /**
   * Determine if the schema exists on disk.
   *
   * @returns True if the schema file exists on disk.
   */
  hasSchemaOnDisk() {
    return fs.access(this.helper.paths.schema).then(() => true).catch(() => false);
  }
  /**
   * Load the schema either from disk or by downloading it.
   *
   * @param forceDownload - Forces downloading the schema.
   */
  async loadSchema(opts) {
    if (opts?.forceDisk) {
      this.schemaContent = await this.loadSchemaFromDisk();
    } else if (this.helper.options.downloadSchema || opts?.forceDownload) {
      this.schemaContent = await this.downloadSchema();
    } else {
      this.schemaContent = await this.loadSchemaFromDisk();
    }
    this.schema = await loadSchema(this.schemaContent, {
      loaders: []
    });
  }
  /**
   * Get the schema.
   *
   * @returns The parsed GraphQL schema object.
   */
  getSchema() {
    if (!this.schema) {
      throw new Error("Failed to load schema.");
    }
    return this.schema;
  }
}

const unicode = isUnicodeSupported();
const s = (c, fallback) => unicode ? c : fallback;
const S_BAR = s("\u2502", "|");
const S_STEP_ACTIVE = s("\u25C6", "*");
const S_STEP_CANCEL = s("\u25A0", "x");
const S_STEP_ERROR = s("\u25B2", "x");
const S_STEP_SUBMIT = s("\u25C7", "o");
const S_RADIO_ACTIVE = s("\u25CF", ">");
const S_RADIO_INACTIVE = s("\u25CB", " ");
const S_BAR_END = s("\u2514", "\u2014");
const symbol = (state) => {
  switch (state) {
    case "initial":
    case "active":
      return color.cyan(S_STEP_ACTIVE);
    case "cancel":
      return color.red(S_STEP_CANCEL);
    case "error":
      return color.yellow(S_STEP_ERROR);
    case "submit":
      return color.green(S_STEP_SUBMIT);
  }
};
class ConsolePrompt {
  abortController = null;
  confirm(message) {
    this.abort();
    this.abortController = new AbortController();
    const active = "Yes";
    const inactive = "No";
    return new ConfirmPrompt({
      active,
      inactive,
      initialValue: true,
      signal: this.abortController.signal,
      render() {
        const title = `${color.gray(S_BAR)}
${symbol(this.state)}  ${message}
`;
        const value = this.value ? active : inactive;
        switch (this.state) {
          case "submit":
            return `${title}${color.gray(S_BAR)}  ${color.dim(value)}`;
          case "cancel":
            return `${title}${color.gray(S_BAR)}  ${color.strikethrough(
              color.dim(value)
            )}
${color.gray(S_BAR)}`;
          default: {
            return `${title}${color.cyan(S_BAR)}  ${this.value ? `${color.green(S_RADIO_ACTIVE)} ${active}` : `${color.dim(S_RADIO_INACTIVE)} ${color.dim(active)}`} ${color.dim("/")} ${!this.value ? `${color.green(S_RADIO_ACTIVE)} ${inactive}` : `${color.dim(S_RADIO_INACTIVE)} ${color.dim(inactive)}`}
${color.cyan(S_BAR_END)}
`;
          }
        }
      }
    }).prompt().then((v) => {
      const result = v;
      if (result === true) {
        return "yes";
      } else if (result === false) {
        return "no";
      }
      return "cancel";
    });
  }
  abort() {
    if (this.abortController) {
      this.abortController.abort();
      this.abortController = null;
    }
  }
}

class ModuleHelper {
  constructor(nuxt, moduleUrl, options) {
    this.nuxt = nuxt;
    const isModuleBuild = process.env.MODULE_BUILD === "true" && nuxt.options._prepare;
    const mergedOptions = defu({}, options, defaultOptions);
    if (!mergedOptions.autoImportPatterns) {
      mergedOptions.autoImportPatterns = [
        "~~/**/*.{gql,graphql}",
        "!node_modules"
      ];
    }
    if (isModuleBuild) {
      mergedOptions.graphqlEndpoint = "http://localhost";
      mergedOptions.downloadSchema = false;
      mergedOptions.schemaPath = "~~/schema.graphql";
      mergedOptions.autoImportPatterns = [
        "~~/playground/**/*.{gql,graphql}",
        "!node_modules"
      ];
    }
    mergedOptions.autoImportPatterns = (mergedOptions.autoImportPatterns || []).map((pattern) => {
      return resolveAlias(pattern);
    });
    this.options = mergedOptions;
    if (!nuxt.options._prepare) {
      validateOptions(this.options);
    }
    this.isDev = nuxt.options.dev;
    this.resolvers = {
      module: createResolver(moduleUrl),
      server: createResolver(nuxt.options.serverDir),
      src: createResolver(nuxt.options.srcDir),
      app: createResolver(nuxt.options.dir.app),
      root: createResolver(nuxt.options.rootDir)
    };
    this.paths = {
      runtimeTypes: "",
      root: nuxt.options.rootDir,
      nuxtConfig: this.resolvers.root.resolve("nuxt.config.ts"),
      serverDir: nuxt.options.serverDir,
      schema: this.resolvers.root.resolve(
        resolveAlias(this.options.schemaPath)
      ),
      serverOptions: "",
      clientOptions: this.findClientOptions(),
      moduleBuildDir: nuxt.options.buildDir + "/nuxt-graphql-middleware",
      moduleTypesDir: nuxt.options.buildDir + "/graphql-operations"
    };
    this.paths.runtimeTypes = this.toModuleBuildRelative(
      this.resolvers.module.resolve("./runtime/types.ts")
    );
    this.paths.serverOptions = this.findServerOptions();
  }
  resolvers;
  paths;
  isDev;
  options;
  prompt = new ConsolePrompt();
  nitroExternals = [];
  /**
   * Find the path to the graphqlMiddleware.serverOptions.ts file.
   */
  findServerOptions() {
    const newPath = this.resolvers.server.resolve(
      "graphqlMiddleware.serverOptions"
    );
    const serverPath = fileExists(newPath);
    if (serverPath) {
      return serverPath;
    }
    const candidates = [
      this.resolvers.root.resolve("graphqlMiddleware.serverOptions"),
      this.resolvers.root.resolve("app/graphqlMiddleware.serverOptions"),
      this.resolvers.src.resolve("graphqlMiddleware.serverOptions")
    ];
    for (let i = 0; i < candidates.length; i++) {
      const path = candidates[i];
      const filePath = fileExists(path);
      if (filePath) {
        throw new Error(
          `The graphqlMiddleware.serverOptions file should be placed in Nuxt's <serverDir> ("${this.paths.serverDir}/graphqlMiddleware.serverOptions.ts").`
        );
      }
    }
    logger.info("No graphqlMiddleware.serverOptions file found.");
    return null;
  }
  findClientOptions() {
    const clientOptionsPath = this.resolvers.app.resolve(
      "graphqlMiddleware.clientOptions"
    );
    if (fileExists(clientOptionsPath)) {
      return clientOptionsPath;
    }
    return null;
  }
  /**
   * Transform the path relative to the module's build directory.
   *
   * @param path - The absolute path.
   *
   * @returns The path relative to the module's build directory.
   */
  toModuleBuildRelative(path) {
    return relative(this.paths.moduleBuildDir, path);
  }
  /**
   * Transform the path relative to the Nuxt build directory.
   *
   * @param path - The absolute path.
   *
   * @returns The path relative to the module's build directory.
   */
  toBuildRelative(path) {
    return relative(this.nuxt.options.buildDir, path);
  }
  /**
   * Get all file paths that match the import patterns.
   */
  async getImportPatternFiles() {
    return resolveFiles(
      this.nuxt.options.srcDir,
      this.options.autoImportPatterns,
      {
        followSymbolicLinks: false
      }
    );
  }
  matchesImportPattern(filePath) {
    return micromatch.isMatch(filePath, this.options.autoImportPatterns);
  }
  addAlias(name, path) {
    this.nuxt.options.alias[name] = path;
    const pathFromName = `./${name.substring(1)}`;
    this.nuxt.options.nitro.typescript ||= {};
    this.nuxt.options.nitro.typescript.tsConfig ||= {};
    this.nuxt.options.nitro.typescript.tsConfig.compilerOptions ||= {};
    this.nuxt.options.nitro.typescript.tsConfig.compilerOptions.paths ||= {};
    this.nuxt.options.nitro.typescript.tsConfig.compilerOptions.paths[name] = [
      pathFromName
    ];
    this.nuxt.options.nitro.typescript.tsConfig.compilerOptions.paths[name + "/*"] = [pathFromName + "/*"];
    this.nuxt.options.typescript.tsConfig ||= {};
    this.nuxt.options.typescript.tsConfig.compilerOptions ||= {};
    this.nuxt.options.typescript.tsConfig.compilerOptions.paths ||= {};
    this.nuxt.options.typescript.tsConfig.compilerOptions.paths[name] = [
      pathFromName
    ];
    this.nuxt.options.typescript.tsConfig.compilerOptions.paths[name + "/*"] = [
      pathFromName + "/*"
    ];
    this.inlineNitroExternals(name);
  }
  inlineNitroExternals(arg) {
    const path = typeof arg === "string" ? arg : arg.dst;
    this.nitroExternals.push(path);
    this.transpile(path);
  }
  transpile(path) {
    this.nuxt.options.build.transpile.push(path);
  }
  applyBuildConfig() {
    this.nuxt.options.nitro.externals ||= {};
    this.nuxt.options.nitro.externals.inline ||= [];
    this.nuxt.options.nitro.externals.inline.push(...this.nitroExternals);
  }
  addTemplate(template) {
    if (template.build) {
      const content = template.build(this).trim();
      addTemplate({
        filename: template.options.path + ".js",
        write: true,
        getContents: () => content
      });
    }
    if (template.buildTypes) {
      const content = template.buildTypes(this).trim();
      const filename = template.options.path + ".d.ts";
      addTypeTemplate({
        filename,
        write: true,
        getContents: () => content
      });
    }
  }
  addPlugin(path) {
    addPlugin(this.resolvers.module.resolve(path), {
      append: false
    });
  }
  addServerHandler(name, path, method) {
    addServerHandler({
      handler: this.resolvers.module.resolve("./runtime/server/api/" + name),
      route: this.options.serverApiPrefix + path,
      method
    });
  }
  addComposable(name) {
    addImports({
      from: this.resolvers.module.resolve("./runtime/composables/" + name),
      name
    });
  }
  addServerUtil(name) {
    addServerImports([
      {
        from: this.resolvers.module.resolve("./runtime/server/utils/" + name),
        name
      }
    ]);
  }
}

function defineGeneratorTemplate(options, build, buildTypes) {
  return {
    type: "generator",
    options,
    build,
    buildTypes
  };
}
function defineStaticTemplate(options, build, buildTypes) {
  return {
    type: "static",
    options,
    build,
    buildTypes
  };
}

const ClientOptions = defineStaticTemplate(
  { path: "nuxt-graphql-middleware/client-options" },
  (helper) => {
    if (helper.paths.clientOptions) {
      const pathRelative = helper.toModuleBuildRelative(
        helper.paths.clientOptions
      );
      return `import clientOptions from '${pathRelative}'
export { clientOptions }
`;
    }
    return `export const clientOptions = {}`;
  },
  (helper) => {
    if (helper.paths.clientOptions) {
      const pathRelative = helper.toModuleBuildRelative(
        helper.paths.clientOptions
      );
      return `import type { GraphqlClientOptions } from '${helper.paths.runtimeTypes}'
import { clientOptions } from '${pathRelative}'

export type GraphqlClientContext = typeof clientOptions extends GraphqlClientOptions<infer R> ? R : {}

export { clientOptions }`;
    }
    return `
import type { GraphqlClientOptions } from '${helper.paths.runtimeTypes}'
export const clientOptions: GraphqlClientOptions

export type GraphqlClientContext = {}
`;
  }
);

const Documents = defineGeneratorTemplate(
  { path: "nuxt-graphql-middleware/documents", virtual: true },
  (output, helper) => {
    return output.getOperationsFile({
      exportName: "documents",
      minify: !helper.isDev
    }).getSource();
  },
  () => {
    return `
import type { Query, Mutation } from './operation-types'

declare module '#nuxt-graphql-middleware/documents' {
  export type Documents = {
    query: Record<keyof Query, string>
    mutation: Record<keyof Mutation, string>
  }
  export const documents: Documents
}`;
  }
);

const GraphqlConfig = defineStaticTemplate(
  { path: "nuxt-graphql-middleware/graphql.config" },
  (helper) => {
    const patterns = helper.options.autoImportPatterns || [];
    const configPath = helper.resolvers.root.resolve(
      (helper.options.graphqlConfigFilePath || "").replace(
        "/graphql.config.ts",
        ""
      )
    );
    const schemaPath = "./" + relative(configPath, helper.paths.schema);
    const documents = patterns.filter((v) => !v.includes("!")).map((pattern) => {
      return "./" + relative(configPath, helper.resolvers.root.resolve(pattern));
    });
    return `const schema = ${JSON.stringify(schemaPath)}

const documents = ${JSON.stringify(documents, null, 2)};

const config = {
  schema,
  documents,
}

export default config
`;
  },
  () => {
    return `
import type { IGraphQLProject } from 'graphql-config'

type WithRequired<T, K extends keyof T> = T & { [P in K]-?: T[P] };

const config: WithRequired<IGraphQLProject, 'schema' | 'documents'>;

export default config;
`;
  }
);

const Helpers = defineStaticTemplate(
  { path: "nuxt-graphql-middleware/helpers" },
  (helper) => {
    return `export const serverApiPrefix = '${helper.options.serverApiPrefix}'
export function getEndpoint(operation, operationName) {
  return serverApiPrefix + '/' + operation + '/' + operationName
}
`;
  },
  () => {
    return `export const serverApiPrefix: string;
export function getEndpoint(operation: string, operationName: string): string`;
  }
);

const NitroTypes = defineGeneratorTemplate(
  { path: "nuxt-graphql-middleware/nitro" },
  null,
  (output, helper) => {
    const operations = output.getCollectedOperations();
    const serverApiPrefix = helper.options.serverApiPrefix;
    const endpoints = [];
    const imports = [];
    for (const operation of operations) {
      imports.push(operation.typeName);
      const method = operation.operationType === OperationTypeNode.QUERY ? "get" : "post";
      endpoints.push(
        `    '${serverApiPrefix}/${operation.operationType}/${operation.graphqlName}': {
      '${method}': GraphqlResponse<${operation.typeName}>
    }`
      );
    }
    return `import type { GraphqlResponse } from './response'
import type {
  ${imports.sort().join(",\n  ")}
} from './../graphql-operations'

declare module 'nitropack/types' {
  interface InternalApi {
${endpoints.sort().join("\n")}
  }
}`;
  }
);

const OperationTypesAll = defineGeneratorTemplate(
  { path: "nuxt-graphql-middleware/operation-types" },
  () => `export {}`,
  (output) => {
    return output.getOperationTypesFile({
      importFrom: "./../graphql-operations"
    }).getSource();
  }
);

const Operations = defineGeneratorTemplate(
  { path: "graphql-operations/index" },
  (output) => {
    const typesFile = output.getOperations("js");
    return typesFile.getSource();
  },
  (output) => {
    const typesFile = output.getOperations("d.ts");
    return typesFile.getSource();
  }
);

const Response = defineGeneratorTemplate(
  { path: "nuxt-graphql-middleware/response" },
  null,
  (output, helper) => {
    const operations = output.getCollectedOperations();
    const allTypes = operations.map((v) => v.typeName).sort();
    return `import type {
  ${allTypes.join(",\n  ")}
} from './../graphql-operations'
import type { GraphqlResponseAdditions } from './server-options'
import type { GraphqlServerResponse } from '${helper.paths.runtimeTypes}'

declare module '#nuxt-graphql-middleware/response' {
  export type GraphqlMiddlewareResponseUnion =
    | ${allTypes.join("\n  | ") || "never"}

  export type GraphqlResponse<T> = GraphqlServerResponse<T> & GraphqlResponseAdditions
  export type GraphqlResponseTyped = GraphqlResponse<GraphqlMiddlewareResponseUnion>
}`;
  }
);

const ServerOptions = defineStaticTemplate(
  { path: "nuxt-graphql-middleware/server-options" },
  (helper) => {
    const resolvedPathRelative = helper.paths.serverOptions ? helper.toModuleBuildRelative(helper.paths.serverOptions) : null;
    const serverOptionsLine = resolvedPathRelative ? `import serverOptions from '${resolvedPathRelative}'` : `const serverOptions = {}`;
    return `
${serverOptionsLine}
export { serverOptions }
`;
  },
  (helper) => {
    const resolvedPathRelative = helper.paths.serverOptions ? helper.toModuleBuildRelative(helper.paths.serverOptions) : null;
    const serverOptionsLineTypes = resolvedPathRelative ? `import serverOptions from '${resolvedPathRelative}'` : `const serverOptions: GraphqlMiddlewareServerOptions = {}`;
    return `
import type { GraphqlMiddlewareServerOptions } from '${helper.paths.runtimeTypes}'
${serverOptionsLineTypes}

export type GraphqlResponseAdditions =
  typeof serverOptions extends GraphqlMiddlewareServerOptions<infer R, any, any> ? R : {}

export { serverOptions }`;
  }
);

const Sources = defineGeneratorTemplate(
  { path: "nuxt-graphql-middleware/sources" },
  (output, helper) => {
    const operations = output.getCollectedOperations();
    const srcDir = helper.paths.root;
    const lines = [];
    for (const operation of operations) {
      const filePath = relative(srcDir, operation.filePath);
      lines.push(
        `${operation.operationType}_${operation.graphqlName}: '${filePath}',`
      );
    }
    return `
export const operationSources = {
  ${lines.join("\n  ")}
}
`;
  },
  () => {
    return `export const operationSources: Record<string, string>`;
  }
);

const TEMPLATES = [
  ClientOptions,
  Documents,
  GraphqlConfig,
  Helpers,
  NitroTypes,
  OperationTypesAll,
  Operations,
  Response,
  ServerOptions,
  Sources
];

const DEVTOOLS_UI_ROUTE = "/__nuxt-graphql-middleware";
const DEVTOOLS_UI_LOCAL_PORT = 3300;
function setupDevToolsUI(nuxt, clientPath) {
  const isProductionBuild = existsSync$1(clientPath);
  if (isProductionBuild) {
    nuxt.hook("vite:serverCreated", async (server) => {
      const sirv = await import('sirv').then((r) => r.default || r);
      server.middlewares.use(
        DEVTOOLS_UI_ROUTE,
        sirv(clientPath, { dev: true, single: true })
      );
    });
  } else {
    nuxt.hook("vite:extendConfig", (config) => {
      config.server = config.server || {};
      config.server.proxy = config.server.proxy || {};
      config.server.proxy[DEVTOOLS_UI_ROUTE] = {
        target: "http://localhost:" + DEVTOOLS_UI_LOCAL_PORT + DEVTOOLS_UI_ROUTE,
        changeOrigin: true,
        followRedirects: true,
        rewrite: (path) => path.replace(DEVTOOLS_UI_ROUTE, "")
      };
    });
  }
  nuxt.hook("devtools:customTabs", (tabs) => {
    tabs.push({
      // unique identifier
      name: "nuxt-graphql-middleware",
      // title to display in the tab
      title: "GraphQL Middleware",
      // any icon from Iconify, or a URL to an image
      icon: "akar-icons:graphql-fill",
      // iframe view
      view: {
        type: "iframe",
        src: DEVTOOLS_UI_ROUTE
      }
    });
  });
}

const RPC_NAMESPACE = "nuxt-graphql-middleware";
class DevModeHandler {
  constructor(nuxt, schemaProvider, collector, helper) {
    this.nuxt = nuxt;
    this.schemaProvider = schemaProvider;
    this.collector = collector;
    this.helper = helper;
  }
  devToolsRpc = null;
  nitro = null;
  viteWebSocket = null;
  operationsToReload = /* @__PURE__ */ new Set();
  init() {
    this.nuxt.hooks.hookOnce("ready", this.onReady.bind(this));
    this.nuxt.hooks.hookOnce(
      "vite:serverCreated",
      this.onViteServerCreated.bind(this)
    );
    this.nuxt.hook("builder:watch", this.onBuilderWatch.bind(this));
    if (this.helper.options.devtools) {
      const clientPath = this.helper.resolvers.module.resolve("./client");
      setupDevToolsUI(this.nuxt, clientPath);
      onDevToolsInitialized(() => {
        this.devToolsRpc = extendServerRpc(
          RPC_NAMESPACE,
          {
            // register server RPC functions
            getModuleOptions: () => {
              return this.helper.options;
            },
            getDocuments: () => {
              return [...this.collector.rpcItems.values()];
            }
          }
        );
      });
    }
  }
  onReady() {
    this.nitro = useNitro();
    this.nitro.hooks.hook("compiled", this.onNitroCompiled.bind(this));
  }
  async onBuilderWatch(event, pathAbsolute) {
    if (pathAbsolute === this.helper.paths.schema) {
      return;
    }
    if (!pathAbsolute.match(/\.(gql|graphql)$/)) {
      return;
    }
    this.helper.prompt.abort();
    const { hasChanged, affectedOperations, error } = await this.collector.handleWatchEvent(event, pathAbsolute);
    if (error) {
      this.sendError(error);
      await this.helper.prompt.confirm("Do you want to download and update the GraphQL schema?").then(async (shouldReload) => {
        if (shouldReload !== "yes") {
          return;
        }
        try {
          await this.schemaProvider.loadSchema({ forceDownload: true });
          await this.collector.updateSchema(this.schemaProvider.getSchema());
        } catch (e) {
          logger.error(e);
        }
      });
      return;
    }
    if (!hasChanged) {
      return;
    }
    if (this.nitro) {
      await this.nitro.hooks.callHook("rollup:reload");
    }
    if (affectedOperations.length) {
      affectedOperations.forEach(
        (operation) => this.operationsToReload.add(operation)
      );
    }
    if (this.devToolsRpc) {
      try {
        this.devToolsRpc.broadcast.documentsUpdated([
          ...this.collector.rpcItems.values()
        ]);
      } catch {
        logger.info(
          "Failed to update GraphQL documents in dev tools. The documents might be stale."
        );
      }
    }
  }
  onViteServerCreated(server) {
    this.viteWebSocket = server.ws;
  }
  sendError(error) {
    if (!this.viteWebSocket) {
      return;
    }
    this.viteWebSocket.send({
      type: "error",
      err: {
        message: error.message,
        stack: ""
      }
    });
  }
  onNitroCompiled() {
    if (!this.operationsToReload.size) {
      return;
    }
    const operations = [...this.operationsToReload.values()];
    this.operationsToReload.clear();
    if (!this.viteWebSocket) {
      return;
    }
    this.viteWebSocket.send({
      type: "custom",
      event: "nuxt-graphql-middleware:reload",
      data: { operations }
    });
  }
}

const module = defineNuxtModule({
  meta: {
    name,
    configKey: "graphqlMiddleware",
    version,
    compatibility: {
      nuxt: ">=3.15.0"
    }
  },
  defaults: defaultOptions,
  async setup(passedOptions, nuxt) {
    const helper = new ModuleHelper(nuxt, import.meta.url, passedOptions);
    const schemaProvider = new SchemaProvider(helper);
    await schemaProvider.init();
    const collector = new Collector(schemaProvider.getSchema(), helper);
    nuxt.options.appConfig.graphqlMiddleware = {
      clientCacheEnabled: !!helper.options.clientCache?.enabled,
      clientCacheMaxSize: helper.options.clientCache?.maxSize ?? 100
    };
    nuxt.options.runtimeConfig.graphqlMiddleware = {
      graphqlEndpoint: helper.options.graphqlEndpoint || ""
    };
    helper.transpile(fileURLToPath(new URL("./runtime", import.meta.url)));
    helper.inlineNitroExternals(helper.resolvers.module.resolve("./runtime"));
    helper.inlineNitroExternals(helper.paths.moduleBuildDir);
    helper.inlineNitroExternals(helper.paths.moduleTypesDir);
    helper.addAlias("#nuxt-graphql-middleware", helper.paths.moduleBuildDir);
    helper.addAlias("#graphql-operations", helper.paths.moduleTypesDir);
    helper.addPlugin("./runtime/plugins/provideState");
    if (helper.isDev && helper.options.errorOverlay) {
      helper.addPlugin("./runtime/plugins/devMode");
    }
    helper.addServerHandler("query", "/query/:name", "get");
    helper.addServerHandler("mutation", "/mutation/:name", "post");
    if (helper.options.enableFileUploads) {
      helper.addServerHandler("upload", "/upload/:name", "post");
    }
    if (helper.isDev) {
      helper.addServerHandler("debug", "/debug", "get");
    }
    if (helper.options.includeComposables) {
      helper.addComposable("useGraphqlQuery");
      helper.addComposable("useGraphqlMutation");
      helper.addComposable("useGraphqlState");
      helper.addComposable("useAsyncGraphqlQuery");
      if (helper.options.enableFileUploads) {
        helper.addComposable("useGraphqlUploadMutation");
      }
      helper.addServerUtil("useGraphqlQuery");
      helper.addServerUtil("useGraphqlMutation");
      helper.addServerUtil("doGraphqlRequest");
    }
    TEMPLATES.forEach((template) => {
      if (template.type === "static") {
        helper.addTemplate(template);
      } else {
        collector.addTemplate(template);
      }
    });
    helper.applyBuildConfig();
    await collector.init();
    if (!helper.isDev) {
      return;
    }
    const devModeHandler = new DevModeHandler(
      nuxt,
      schemaProvider,
      collector,
      helper
    );
    devModeHandler.init();
  }
});

export { module as default };
