import {
  defineEventHandler,
  readMultipartFormData,
  getQuery,
  getRouterParam
} from "h3";
import {
  getEndpoint,
  getFetchOptions,
  onServerResponse,
  onServerError,
  throwError,
  extractRequestContext,
  isValidMutation
} from "./../helpers/index.js";
import { documents } from "#nuxt-graphql-middleware/documents";
import { serverOptions } from "#nuxt-graphql-middleware/server-options";
import { useRuntimeConfig } from "#imports";
import { GraphqlMiddlewareOperation } from "../../settings/index.js";
function parseMultipart(data) {
  const files = data.filter((v) => !!v.filename);
  const variablesData = data.find((v) => v.name === "variables")?.data?.toString();
  if (!variablesData) {
    throw new Error("Missing variables in form data.");
  }
  const variables = JSON.parse(variablesData);
  const map = data.find((v) => v.name === "map")?.data.toString() || "{}";
  return { files, map, variables };
}
export default defineEventHandler(async (event) => {
  const operation = GraphqlMiddlewareOperation.Mutation;
  const operationName = getRouterParam(event, "name");
  if (!isValidMutation(operationName)) {
    return throwError("Invalid mutation name.");
  }
  const operationDocument = documents.mutation[operationName];
  const multiPartData = await readMultipartFormData(event);
  if (!multiPartData) {
    return throwError("Failed to read multi part data.");
  }
  const { variables, map, files } = parseMultipart(multiPartData);
  const formData = new FormData();
  formData.append(
    "operations",
    JSON.stringify({
      query: operationDocument,
      variables,
      operationName
    })
  );
  formData.append("map", map);
  for (const file of files) {
    if (!file.name) {
      return throwError("Missing name for file: " + file.filename);
    }
    const blob = new Blob([file.data]);
    formData.append(file.name, blob, file.filename);
  }
  const queryParams = getQuery(event);
  const context = extractRequestContext(queryParams);
  if (serverOptions.doGraphqlRequest) {
    return serverOptions.doGraphqlRequest({
      event,
      operation,
      operationName,
      operationDocument,
      variables,
      formData,
      context
    });
  }
  const runtimeConfig = useRuntimeConfig().graphqlMiddleware;
  const endpoint = await getEndpoint(
    runtimeConfig,
    serverOptions,
    event,
    operation,
    operationName,
    context
  );
  const fetchOptions = await getFetchOptions(
    serverOptions,
    event,
    operation,
    operationName,
    context
  );
  return $fetch.raw(endpoint, {
    ...fetchOptions,
    method: "POST",
    body: formData
  }).then((response) => {
    return onServerResponse(
      serverOptions,
      event,
      response,
      operation,
      operationName,
      context
    );
  }).catch((error) => {
    return onServerError(
      serverOptions,
      event,
      error,
      operation,
      operationName,
      context
    );
  });
});
